@@ -1,11 +1,13 @@
 from __future__ import unicode_literals
+import argon2
 import base64
 import os.path
 import argparse
+import secrets
 import sys
 import uuid
 from operator import itemgetter
-
+from tempfile import NamedTemporaryFile
 from pluggy import HookimplMarker, PluginManager
 import py
 from devpi_common.types import cached_property
@@ -26,6 +28,10 @@ hookimpl = HookimplMarker("devpiserver")
 DEFAULT_MIRROR_CACHE_EXPIRY = 1800
 DEFAULT_PROXY_TIMEOUT = 30
 DEFAULT_REQUEST_TIMEOUT = 5
+DEFAULT_FILE_REPLICATION_THREADS = 5
+DEFAULT_ARGON2_MEMORY_COST = 524288
+DEFAULT_ARGON2_PARALLELISM = 8
+DEFAULT_ARGON2_TIME_COST = 16
 
 
 def get_pluginmanager(load_entrypoints=True):
@@ -104,6 +110,14 @@ def add_web_options(parser, pluginmanager):
         default=3141,
         help="port to listen for http requests.")
 
+    parser.addoption(
+        "--listen", type=str, action="append",
+        help="host:port combination to listen to for http requests. "
+             "When using * for host bind to all interfaces. "
+             "Use square brackets for ipv6 like [::1]:8080. "
+             "You can specify more than one host:port combination "
+             "with multiple --listen arguments.")
+
     parser.addoption(
         "--unix-socket", type=str,
         help="path to unix socket to bind to.")
@@ -117,6 +131,18 @@ def add_web_options(parser, pluginmanager):
         default=50,
         help="number of threads to start for serving clients.")
 
+    parser.addoption(
+        "--trusted-proxy", type=str,
+        help="IP address of proxy we trust. See waitress documentation.")
+
+    parser.addoption(
+        "--trusted-proxy-count", type=int,
+        help="how many proxies we trust when chained. See waitress documentation.")
+
+    parser.addoption(
+        "--trusted-proxy-headers", type=str,
+        help="headers to trust from proxy. See waitress documentation.")
+
     parser.addoption(
         "--max-request-body-size", type=int,
         default=1073741824,
@@ -178,6 +204,11 @@ def add_replica_options(parser, pluginmanager):
              "(EXPERIMENTAL)",
         default=None)
 
+    parser.addoption(
+        "--file-replication-threads", type=int, metavar="NUM",
+        default=DEFAULT_FILE_REPLICATION_THREADS,
+        help="number of threads for file download from master")
+
     parser.addoption(
         "--proxy-timeout", type=int, metavar="NUM",
         default=DEFAULT_PROXY_TIMEOUT,
@@ -223,14 +254,7 @@ def add_storage_options(parser, pluginmanager):
              "average. So by default about 10MB are used.")
 
 
-def add_init_options(parser, pluginmanager, standalone=True):
-    if not standalone:
-        parser.addoption(
-            "--init", action="store_true",
-            help="(DEPRECATED, use devpi-init command) initialize "
-                 "devpi-server state in an empty directory "
-                 "(also see --serverdir)")
-
+def add_init_options(parser, pluginmanager):
     parser.addoption(
         "--no-root-pypi", action="store_true",
         help="don't create root/pypi on server initialization.")
@@ -247,33 +271,13 @@ def add_init_options(parser, pluginmanager, standalone=True):
              "exist.")
 
 
-def add_export_options(parser, pluginmanager, standalone=True):
-    if not standalone:
-        parser.addoption(
-            "--export", type=str, metavar="PATH",
-            help="(DEPRECATED, use devpi-passwd command) export "
-                 "devpi-server database state into PATH. "
-                 "This will export all users, indices, release files "
-                 "(except for mirrors), test results and documentation.")
-
+def add_export_options(parser, pluginmanager):
     parser.addoption(
         "--include-mirrored-files", action="store_true",
         help="include downloaded files from mirror indexes in dump.")
 
 
-def add_import_options(parser, pluginmanager, standalone=True):
-    if not standalone:
-        parser.addoption(
-            "--import", type=str, metavar="PATH",
-            dest="import_",
-            help="(DEPRECATED, use devpi-import command) import "
-                 "devpi-server database from PATH where PATH "
-                 "is a directory which was created by a "
-                 "'devpi-server --export PATH' operation, "
-                 "using the same or an earlier devpi-server version. "
-                 "Note that you can only import into a fresh server "
-                 "state directory (positional argument to devpi-server).")
-
+def add_import_options(parser, pluginmanager):
     parser.addoption(
         "--skip-import-type", action="append", metavar="TYPE",
         help="skip the given index type during import. "
@@ -282,7 +286,7 @@ def add_import_options(parser, pluginmanager, standalone=True):
     parser.addoption(
         "--no-events", action="store_false",
         default=True, dest="wait_for_events",
-        help="no events will be run during import, instead they are"
+        help="no events will be run during import, instead they are "
              "postponed to run on server start. This allows much faster "
              "start of the server after import, when devpi-web is used. "
              "When you start the server after the import, the search "
@@ -290,20 +294,40 @@ def add_import_options(parser, pluginmanager, standalone=True):
              "server has caught up with all events.")
 
 
-def add_deploy_options(parser, pluginmanager):
-    parser.addoption(
-        "--gen-config", dest="genconfig", action="store_true",
-        help="(DEPRECATED, use devpi-gen-config command) "
-             "generate example config files for "
-             "nginx/supervisor/crontab/systemd/launchd/windows-service, "
-             "taking other passed options into account "
-             "(e.g. port, host, etc.)")
-
+def add_secretfile_option(parser, pluginmanager):
     parser.addoption(
         "--secretfile", type=str, metavar="path",
         help="file containing the server side secret used for user "
              "validation. If not specified, a random secret is "
-             "generated on each start up. ")
+             "generated on each start up.")
+
+
+def add_deploy_options(parser, pluginmanager):
+    add_secretfile_option(parser, pluginmanager)
+
+    parser.addoption(
+        "--argon2-memory-cost", type=int, default=DEFAULT_ARGON2_MEMORY_COST,
+        help="Argon2 memory cost parameter for key derivation. "
+             "This is *not* for the user password hashes. "
+             "There should be no need to touch this setting, "
+             "except you really know what this is about! "
+             "Replicas need to use the same parameters as the master.")
+
+    parser.addoption(
+        "--argon2-parallelism", type=int, default=DEFAULT_ARGON2_PARALLELISM,
+        help="Argon2 parallelism parameter for key derivation. "
+             "This is *not* for the user password hashes. "
+             "There should be no need to touch this setting, "
+             "except you really know what this is about! "
+             "Replicas need to use the same parameters as the master.")
+
+    parser.addoption(
+        "--argon2-time-cost", type=int, default=DEFAULT_ARGON2_TIME_COST,
+        help="Argon2 time cost parameter for key derivation. "
+             "This is *not* for the user password hashes. "
+             "There should be no need to touch this setting, "
+             "except you really know what this is about! "
+             "Replicas need to use the same parameters as the master.")
 
     parser.addoption(
         "--requests-only", action="store_true",
@@ -358,15 +382,6 @@ def addoptions(parser, pluginmanager):
     add_storage_options(
         parser.addgroup("storage options"),
         pluginmanager)
-    add_init_options(
-        parser.addgroup("initialization options"),
-        pluginmanager, standalone=False)
-    add_import_options(
-        parser.addgroup("serverstate import options"),
-        pluginmanager, standalone=False)
-    add_export_options(
-        parser.addgroup("serverstate export options"),
-        pluginmanager, standalone=False)
     add_deploy_options(
         parser.addgroup("deployment options"),
         pluginmanager)
@@ -374,18 +389,6 @@ def addoptions(parser, pluginmanager):
         parser.addgroup("permission options"),
         pluginmanager)
 
-    bg = parser.addgroup(
-        "background server (DEPRECATED, see --gen-config to use a process "
-        "manager from your OS)")
-    bg.addoption("--start", action="store_true",
-            help="start the background devpi-server")
-    bg.addoption("--stop", action="store_true",
-            help="stop the background devpi-server")
-    bg.addoption("--status", action="store_true",
-            help="show status of background devpi-server")
-    bg.addoption("--log", action="store_true",
-            help="show logfile content of background server")
-
 
 def try_argcomplete(parser):
     try:
@@ -557,6 +560,10 @@ class MyArgumentParser(argparse.ArgumentParser):
         return grp
 
 
+def new_secret():
+    return base64.b64encode(secrets.token_bytes(32))
+
+
 class ConfigurationError(Exception):
     """ incorrect configuration or environment settings. """
 
@@ -566,8 +573,57 @@ class Config(object):
         self.args = args
         self.pluginmanager = pluginmanager
         self.hook = pluginmanager.hook
-        self.serverdir = py.path.local(os.path.expanduser(self.args.serverdir))
-        self.path_nodeinfo = self.serverdir.join(".nodeinfo")
+        self._key_cache = {}
+
+    @cached_property
+    def waitress_info(self):
+        from .main import fatal
+        host = self.args.host
+        port = self.args.port
+        default_host_port = (host == 'localhost') and (port == 3141)
+        addresses = []
+        kwargs = dict(
+            threads=self.args.threads,
+            max_request_body_size=self.args.max_request_body_size)
+        unix_socket = self.args.unix_socket
+        if unix_socket is not None:
+            kwargs['unix_socket'] = unix_socket
+            if self.args.unix_socket_perms is not None:
+                kwargs['unix_socket_perms'] = self.args.unix_socket_perms
+            if default_host_port:
+                host = None
+                port = None
+        if self.args.listen:
+            if not default_host_port:
+                fatal("You can use either --listen or --host/--port, not both together.")
+            host = None
+            port = None
+            for listen in self.args.listen:
+                kwargs.setdefault("listen", []).append(listen)
+                addresses.append("http://%s" % listen)
+        if host or port:
+            kwargs['host'] = host
+            kwargs['port'] = port
+            hostaddr = "http://%s:%s" % (host, port)
+            hostaddr6 = "http://[%s]:%s" % (host, port)
+            addresses.append("%s (might be %s for IPv6)" % (hostaddr, hostaddr6))
+        if "listen" in kwargs:
+            kwargs["listen"] = " ".join(kwargs["listen"])
+        if self.args.trusted_proxy is not None:
+            kwargs["trusted_proxy"] = self.args.trusted_proxy
+        if self.args.trusted_proxy_count is not None:
+            kwargs["trusted_proxy_count"] = self.args.trusted_proxy_count
+        if self.args.trusted_proxy_headers is not None:
+            kwargs["trusted_proxy_headers"] = self.args.trusted_proxy_headers
+        return dict(kwargs=kwargs, addresses=addresses)
+
+    @cached_property
+    def serverdir(self):
+        return py.path.local(os.path.expanduser(self.args.serverdir))
+
+    @cached_property
+    def path_nodeinfo(self):
+        return self.serverdir.join(".nodeinfo")
 
     def init_nodeinfo(self):
         log.info("Loading node info from %s", self.path_nodeinfo)
@@ -612,8 +668,16 @@ class Config(object):
         return {}
 
     def write_nodeinfo(self):
-        self.path_nodeinfo.dirpath().ensure(dir=1)
-        self.path_nodeinfo.write(json.dumps(self.nodeinfo, indent=2))
+        nodeinfo_dir = self.path_nodeinfo.dirpath()
+        nodeinfo_dir.ensure(dir=1)
+        prefix = "-" + self.path_nodeinfo.basename
+        with NamedTemporaryFile(prefix=prefix, delete=False, dir=nodeinfo_dir.strpath) as f:
+            f.write(json.dumps(self.nodeinfo, indent=2).encode('utf-8'))
+        try:
+            os.remove(self.path_nodeinfo.strpath)
+        except FileNotFoundError:
+            pass
+        os.rename(f.name, self.path_nodeinfo.strpath)
         threadlog.info("wrote nodeinfo to: %s", self.path_nodeinfo)
 
     @property
@@ -637,21 +701,45 @@ class Config(object):
             if value.port:
                 netloc = "%s:%s" % (netloc, value.port)
             value = value.replace(netloc=netloc)
+        if auth == (None, None):
+            auth = None
         self.master_auth = auth
         self._master_url = value
 
+    @property
+    def include_mirrored_files(self):
+        return getattr(self.args, 'include_mirrored_files', False)
+
     @property
     def mirror_cache_expiry(self):
         return getattr(self.args, 'mirror_cache_expiry', DEFAULT_MIRROR_CACHE_EXPIRY)
 
+    @property
+    def no_root_pypi(self):
+        return getattr(self.args, 'no_root_pypi', False)
+
     @property
     def offline_mode(self):
         return getattr(self.args, 'offline_mode', False)
 
+    @property
+    def file_replication_threads(self):
+        return getattr(
+            self.args,
+            'file_replication_threads', DEFAULT_FILE_REPLICATION_THREADS)
+
+    @property
+    def hard_links(self):
+        return getattr(self.args, 'hard_links', False)
+
     @property
     def replica_cert(self):
         return getattr(self.args, 'replica_cert', None)
 
+    @property
+    def replica_file_search_path(self):
+        return getattr(self.args, 'replica_file_search_path', None)
+
     @property
     def replica_max_retries(self):
         return getattr(self.args, 'replica_max_retries', None)
@@ -665,12 +753,28 @@ class Config(object):
         return getattr(self.args, 'request_timeout', DEFAULT_REQUEST_TIMEOUT)
 
     @property
+    def root_passwd(self):
+        return getattr(self.args, 'root_passwd', "")
+
+    @property
+    def root_passwd_hash(self):
+        return getattr(self.args, 'root_passwd_hash', None)
+
+    @property
+    def skip_import_type(self):
+        return getattr(self.args, 'skip_import_type', None)
+
+    @cached_property
     def restrict_modify(self):
         rm = self.args.restrict_modify
         if rm is not None:
-            rm = set(x.strip() for x in rm.split(','))
+            rm = frozenset(x.strip() for x in rm.split(','))
         return rm
 
+    @property
+    def wait_for_events(self):
+        return getattr(self.args, 'wait_for_events', False)
+
     def _init_role(self):
         if self.master_url:
             self.nodeinfo["role"] = "replica"
@@ -760,9 +864,7 @@ class Config(object):
 
     def sqlite_file_needed_but_missing(self):
         return (
-            not getattr(self.args, 'init', None)
-            and not getattr(self.args, 'import_', None)
-            and self.storage_info['name'] == 'sqlite'
+            self.storage_info['name'] == 'sqlite'
             and not self.serverdir.join(".sqlite").exists()
         )
 
@@ -780,38 +882,115 @@ class Config(object):
         return py.path.local(
             os.path.expanduser(self.args.secretfile))
 
-    @cached_property
-    def secret(self):
+    def get_validated_secret(self):
         from .main import fatal
+        import stat
+        if not self.secretfile.check(file=True):
+            fatal("The given secret file doesn't exist.")
+        if self.secretfile.stat().mode & stat.S_IRWXO and sys.platform != "win32":
+            fatal("The given secret file is world accessible, the access mode must be user accessible only (0600).")
+        if self.secretfile.stat().mode & stat.S_IRWXG and sys.platform != "win32":
+            fatal("The given secret file is group accessible, the access mode must be user accessible only (0600).")
+        if self.secretfile.dirpath().stat().mode & stat.S_IWGRP and sys.platform != "win32":
+            fatal("The folder of the given secret file is group writable, it must only be writable by the user.")
+        if self.secretfile.dirpath().stat().mode & stat.S_IWOTH and sys.platform != "win32":
+            fatal("The folder of the given secret file is world writable, it must only be writable by the user.")
+        secret = self.secretfile.read_binary()
+        if len(secret) < 32:
+            fatal(
+                "The secret in the given secret file is too short, "
+                "it should be at least 32 characters long.")
+        if len(set(secret)) < 6:
+            fatal(
+                "The secret in the given secret file is too weak, "
+                "it should use less repetition.")
+        return secret
+
+    @cached_property
+    def basesecret(self):
         if self.secretfile is None:
             log.warn(
                 "No secret file provided, creating a new random secret. "
-                "Login tokens issued before are invalidate. "
-                "Use --secretfile option to provide a persistent secret.")
-            return base64.b64encode(os.urandom(32))
-        if not self.secretfile.check(file=True):
-            fatal("The given secret file doesn't exist.")
+                "Login tokens issued before are invalid. "
+                "Use --secretfile option to provide a persistent secret. "
+                "You can create a proper secret with the "
+                "devpi-gen-secret command.")
+            return new_secret()
+        secret = self.get_validated_secret()
         log.info("Using secret file '%s'.", self.secretfile)
-        return self.secretfile.read()
+        return secret
+
+    @property
+    def _secret_parameters(self):
+        # this is a property, so it can easily be changed for tests
+        # see lower_argon2_parameters fixture
+        return argon2.Parameters(
+            type=argon2.low_level.Type.ID,
+            version=argon2.low_level.ARGON2_VERSION,
+            salt_len=16,
+            hash_len=16,
+            time_cost=self.args.argon2_time_cost,
+            memory_cost=self.args.argon2_memory_cost,
+            parallelism=self.args.argon2_parallelism)
+
+    def get_derived_key(self, salt):
+        if salt not in self._key_cache:
+            secret_parameters = self._secret_parameters
+            self._key_cache[salt] = argon2.low_level.hash_secret_raw(
+                self.basesecret,
+                salt,
+                time_cost=secret_parameters.time_cost,
+                memory_cost=secret_parameters.memory_cost,
+                parallelism=secret_parameters.parallelism,
+                hash_len=secret_parameters.hash_len,
+                type=secret_parameters.type,
+                version=secret_parameters.version)
+        return self._key_cache[salt]
+
+    def get_auth_secret(self):
+        return self.get_derived_key(b'devpi-server-auth')
+
+    def get_replica_secret(self):
+        return self.get_derived_key(b'devpi-server-replica')
 
 
 def getpath(path):
     return py.path.local(os.path.expanduser(str(path)))
 
-def render(tw, confname, format=None, **kw):
-    result = render_string(confname, format=format, **kw)
-    return result
 
-def render_string(confname, format=None, **kw):
-    template = confname + ".template"
-    from pkg_resources import resource_string
-    templatestring = resource_string("devpi_server.cfg", template)
-    if not py.builtin._istext(templatestring):
-        templatestring = py.builtin._totext(templatestring, "utf-8")
-
-    kw = dict((x[0], str(x[1])) for x in kw.items())
-    if format is None:
-        result = templatestring.format(**kw)
-    else:
-        result = templatestring % kw
-    return result
+def gensecret():
+    from .log import configure_cli_logging
+    from .log import threadlog as log
+    from .main import Fatal
+    from .main import fatal
+    import stat
+    try:
+        pluginmanager = get_pluginmanager()
+        parser = MyArgumentParser(
+            description="Create a random secret.",
+            add_help=False)
+        add_help_option(parser, pluginmanager)
+        add_configfile_option(parser, pluginmanager)
+        add_secretfile_option(parser, pluginmanager)
+        config = parseoptions(pluginmanager, sys.argv, parser=parser)
+        configure_cli_logging(config.args)
+        if config.args.secretfile is None:
+            fatal("You need to provide a location for the secret file.")
+        if not config.secretfile.exists():
+            with config.secretfile.open("wb") as f:
+                f.write(new_secret())
+            log.info("New secret written to '%s'" % config.secretfile)
+            if sys.platform != "win32":
+                mode = config.secretfile.stat().mode
+                if mode & stat.S_IRWXG or mode & stat.S_IRWXO:
+                    config.secretfile.chmod(0o600)
+                    log.info("Changed file mode to 0600 to adjust access permissions.")
+        else:
+            log.info("Checking existing secret at '%s'" % config.secretfile)
+        # run checks
+        config.get_validated_secret()
+        log.info("Permissions of secret file look good.")
+    except Fatal as e:
+        tw = py.io.TerminalWriter(sys.stderr)
+        tw.line("fatal: %s" % e.args[0], red=True)
+        return 1
