@@ -1,5 +1,6 @@
 from .config import hookimpl
 from .fileutil import BytesForHardlink
+from .interfaces import IStorageConnection2
 from .keyfs_sqlite import BaseConnection
 from .keyfs_sqlite import BaseStorage
 from .log import threadlog, thread_push_log, thread_pop_log
@@ -7,6 +8,7 @@ from .readonly import ReadonlyView
 from .readonly import get_mutable_deepcopy
 from .fileutil import get_write_file_ensure_dir, rename, loads
 from hashlib import sha256
+from zope.interface import implementer
 import errno
 import os
 import re
@@ -44,6 +46,7 @@ class DirtyFile(object):
                 f.write(content)
 
 
+@implementer(IStorageConnection2)
 class Connection(BaseConnection):
     def rollback(self):
         BaseConnection.rollback(self)
@@ -130,6 +133,25 @@ class Connection(BaseConnection):
 class Storage(BaseStorage):
     Connection = Connection
     db_filename = ".sqlite"
+    expected_schema = dict(
+        index=dict(
+            kv_serial_idx="""
+                CREATE INDEX kv_serial_idx ON kv (serial);
+            """),
+        table=dict(
+            changelog="""
+                CREATE TABLE changelog (
+                    serial INTEGER PRIMARY KEY,
+                    data BLOB NOT NULL
+                )
+            """,
+            kv="""
+                CREATE TABLE kv (
+                    key TEXT NOT NULL PRIMARY KEY,
+                    keyname TEXT,
+                    serial INTEGER
+                )
+            """))
 
     def perform_crash_recovery(self):
         # get last changes and verify all renames took place
@@ -140,27 +162,6 @@ class Storage(BaseStorage):
         changes, rel_renames = loads(data)
         check_pending_renames(str(self.basedir), rel_renames)
 
-    def ensure_tables_exist(self):
-        if self.sqlpath.exists():
-            return
-        with self.get_connection(write=True) as conn:
-            threadlog.info("DB: Creating schema")
-            c = conn._sqlconn.cursor()
-            c.execute("""
-                CREATE TABLE kv (
-                    key TEXT NOT NULL PRIMARY KEY,
-                    keyname TEXT,
-                    serial INTEGER
-                )
-            """)
-            c.execute("""
-                CREATE TABLE changelog (
-                    serial INTEGER PRIMARY KEY,
-                    data BLOB NOT NULL
-                )
-            """)
-            conn.commit()
-
 
 @hookimpl
 def devpiserver_storage_backend(settings):
