@@ -780,7 +780,7 @@ class SimpleLinksChanged:
         with self.xom.keyfs.transaction(write=False):
             mirror_stage = self.xom.model.getstage(username, index)
             if mirror_stage and mirror_stage.ixconfig["type"] == "mirror":
-                cache_projectnames = mirror_stage.cache_projectnames.get_inplace()
+                cache_projectnames = mirror_stage.cache_projectnames
                 if cache is None:  # deleted
                     cache_projectnames.discard(project)
                 else:
