@@ -1,12 +1,14 @@
 from pyramid.httpexceptions import HTTPFound
+from devpi_common.types import cached_property
 from devpi_common.types import ensure_unicode
 from devpi_common.validation import normalize_name
 from devpi_server.auth import Auth
-from devpi_server.views import abort, abort_authenticate
+from devpi_server.config import hookimpl
+from devpi_server.views import abort
 from devpi_server.model import UpstreamError
-from pyramid.authentication import CallbackAuthenticationPolicy
-from pyramid.decorator import reify
-from pyramid.security import Allow, Deny, Everyone
+from pyramid.authorization import ACLHelper, Allow, Authenticated, Deny, Everyone
+from pyramid.interfaces import ISecurityPolicy
+from pyramid.request import RequestLocalCache
 
 
 class RootFactory(object):
@@ -15,8 +17,19 @@ class RootFactory(object):
         xom = request.registry['xom']
         self.model = xom.model
         self.restrict_modify = xom.config.restrict_modify
+        self.hook = xom.config.hook
 
-    @reify
+    @cached_property
+    def _user(self):
+        if self.username:
+            return self.model.get_user(self.username)
+
+    @cached_property
+    def _stage(self):
+        if self.username and self.index:
+            return self.model.getstage(self.username, self.index)
+
+    @cached_property
     def matchdict(self):
         result = {}
         if not self.request.matchdict:
@@ -27,33 +40,40 @@ class RootFactory(object):
             result[k] = v
         return result
 
+    @cached_property
     def __acl__(self):
-        acl = []
+        acl = [(Allow, Everyone, 'user_login')]
         if self.restrict_modify is None:
             acl.extend([
                 (Allow, Everyone, 'user_create'),
-                (Allow, 'root', 'user_delete'),
                 (Allow, 'root', 'user_modify'),
                 (Allow, 'root', 'index_create')])
-            if self.username == 'root':
-                acl.append((Deny, Everyone, 'user_delete'))
             if self.username:
+                if self.username != 'root':
+                    acl.extend([
+                        (Allow, 'root', 'user_delete'),
+                        (Allow, self.username, 'user_delete')])
                 acl.extend([
-                    (Allow, self.username, 'user_delete'),
                     (Allow, self.username, 'user_modify'),
                     (Allow, self.username, 'index_create')])
         else:
             for principal in self.restrict_modify:
+                if self.username and self.username != principal:
+                    acl.append((Allow, principal, 'user_delete'))
                 acl.extend([
                     (Allow, principal, 'user_create'),
-                    (Allow, principal, 'user_delete'),
                     (Allow, principal, 'user_modify'),
                     (Allow, principal, 'index_create')])
-        stage = None
-        if self.username and self.index:
-            stage = self.model.getstage(self.username, self.index)
+        stage = self._stage
         if stage:
             acl.extend(stage.__acl__())
+        acl = tuple(acl)
+        all_denials = self.hook.devpiserver_auth_denials(
+            request=self.request, acl=acl, user=self._user, stage=stage)
+        if all_denials:
+            denials = set().union(*all_denials)
+            if denials:
+                acl = tuple((Deny,) + denial for denial in denials) + acl
         return acl
 
     def getstage(self, user, index):
@@ -98,11 +118,11 @@ class RootFactory(object):
             abort(self.request, 404, "no project %r" %(project))
         return res
 
-    @reify
+    @cached_property
     def index(self):
         return self.matchdict.get('index')
 
-    @reify
+    @cached_property
     def project(self):
         project = self.matchdict.get('project')
         if project is None:
@@ -118,7 +138,7 @@ class RootFactory(object):
             raise HTTPFound(location=url)
         return n_project
 
-    @reify
+    @cached_property
     def verified_project(self):
         name = self.project
         try:
@@ -128,77 +148,93 @@ class RootFactory(object):
             abort(self.request, 502, str(e))
         return name
 
-    @reify
+    @cached_property
     def version(self):
         version = self.matchdict.get('version')
         if version is None:
             return
         return ensure_unicode(version)
 
-    @reify
+    @cached_property
     def stage(self):
-        return self.getstage(self.username, self.index)
+        stage = self._stage
+        if not stage:
+            abort(
+                self.request, 404, "The stage %s/%s could not be found." % (
+                    self.username, self.index))
+        return stage
 
-    @reify
+    @cached_property
     def user(self):
-        user = self.model.get_user(self.username)
+        user = self._user
         if not user:
             abort(self.request, 404, "no user %r" % self.username)
         return user
 
-    @reify
+    @cached_property
     def username(self):
         return self.matchdict.get('user')
 
 
-class DevpiAuthenticationPolicy(CallbackAuthenticationPolicy):
+class CredentialsIdentity:
+    def __init__(self, username, groups):
+        self.username = username
+        self.groups = groups
+
+
+class DevpiSecurityPolicy:
     def __init__(self, xom):
         self.realm = "pypi"
-        self.auth = Auth(xom.model, xom.config.secret)
+        self.auth = Auth(xom.model, xom.config.get_auth_secret())
         self.hook = xom.config.hook
+        self.identity_cache = RequestLocalCache(self.load_identity)
 
-    def unauthenticated_userid(self, request):
-        """ The userid parsed from the ``Authorization`` request header."""
-        credentials = self._get_credentials(request)
-        if credentials:
-            return credentials[0]
-
-    def remember(self, request, principal, **kw):
-        """ A no-op. Devpi authentication does not provide a protocol for
-        remembering the user. Credentials are sent on every request.
-        """
+    def remember(self, request, userid, **kw):
+        # A no-op. Devpi authentication does not provide a protocol for
+        # remembering the user. Credentials are sent on every request.
         return []
 
-    def forget(self, request):
-        """ Returns challenge headers. This should be attached to a response
-        to indicate that credentials are required."""
+    def forget(self, request, **kw):
         return [('WWW-Authenticate', 'Basic realm="%s"' % self.realm)]
 
-    def callback(self, username, request):
-        # Username arg is ignored.  Unfortunately _get_credentials winds up
-        # getting called twice when authenticated_userid is called.  Avoiding
-        # that, however, winds up duplicating logic from the superclass.
-        credentials = self._get_credentials(request)
-        if credentials:
-            status, auth_user, groups = self.auth.get_auth_status(credentials)
-            request.log.debug("got auth status %r for user %r" % (status, auth_user))
-            if status == "ok":
-                return [":%s" % g for g in groups]
-            elif status == "reject":
-                abort_authenticate(request, msg="credentials for user '%s' were rejected" % auth_user)
-            elif status == "nouser":
-                abort_authenticate(request, msg="user '%s' does not exist" % auth_user)
-            elif status == "expired":
-                abort_authenticate(request, msg="auth expired for '%s'" % auth_user)
-            raise ValueError("Unknown authentication status: %s" % status)
-
     def _get_credentials(self, request):
         return self.hook.devpiserver_get_credentials(request=request)
 
-    def verify_credentials(self, request):
+    def load_identity(self, request):
         credentials = self._get_credentials(request)
-        if credentials:
-            status = self.auth._validate(*credentials)
-            if status.get("status") == "ok":
-                return True
-        return False
+        return self.hook.devpiserver_get_identity(
+            request=request, credentials=credentials)
+
+    def identity(self, request):
+        return self.identity_cache.get_or_create(request)
+
+    def authenticated_userid(self, request):
+        # defer to the identity logic to determine if the user id logged in
+        # and return None if they are not
+        identity = request.identity
+        if identity is not None:
+            return identity.username
+
+    def permits(self, request, context, permission):
+        # use the identity to build a list of principals, and pass them
+        # to the ACLHelper to determine allowed/denied
+        identity = request.identity
+        principals = set([Everyone])
+        if identity is not None:
+            principals.add(Authenticated)
+            principals.add(identity.username)
+            principals.update(":" + g for g in identity.groups)
+        return ACLHelper().permits(context, principals, permission)
+
+
+@hookimpl(trylast=True)
+def devpiserver_get_identity(request, credentials):
+    if credentials is None:
+        return
+    policy = request.registry.getUtility(ISecurityPolicy)
+    result = policy.auth._get_auth_status(*credentials, request=request)
+    status = result["status"]
+    if status != "ok":
+        return
+    return CredentialsIdentity(
+        credentials[0], result.get("groups", []))
