@@ -18,6 +18,7 @@ from devpi_common.url import URL
 
 import devpi_server
 
+
 def make_export(tmpdir, xom):
     xom.config.init_nodeinfo()
     return do_export(tmpdir, xom)
@@ -54,11 +55,13 @@ def test_not_exists(tmpdir, xom):
     with pytest.raises(Fatal):
         do_import(p, xom)
 
+
 def test_import_wrong_dumpversion(tmpdir, xom):
     tmpdir.join("dataindex.json").write('{"dumpversion": "0"}')
     with pytest.raises(Fatal):
         do_import(tmpdir, xom)
 
+
 def test_empty_export(tmpdir, xom):
     xom.config.init_nodeinfo()
     ret = make_export(tmpdir, xom)
@@ -71,23 +74,15 @@ def test_empty_export(tmpdir, xom):
         make_export(tmpdir, xom)
 
 
-@pytest.mark.parametrize("use_option", (False, True))
-def test_export_empty_serverdir(tmpdir, capfd, monkeypatch, use_option):
+def test_export_empty_serverdir(tmpdir, capfd, monkeypatch):
     from devpi_server.importexport import export
-    from devpi_server.main import main
     empty = tmpdir.join("empty").ensure(dir=True)
     export_dir = tmpdir.join("export")
     monkeypatch.setattr("devpi_server.main.configure_logging", lambda a: None)
-    if use_option:
-        ret = main([
-            "devpi-server",
-            "--serverdir", empty.strpath,
-            "--export", export_dir.strpath])
-    else:
-        ret = export(argv=[
-            "devpi-export",
-            "--serverdir", empty.strpath,
-            export_dir.strpath])
+    ret = export(argv=[
+        "devpi-export",
+        "--serverdir", empty.strpath,
+        export_dir.strpath])
     out, err = capfd.readouterr()
     assert empty.listdir() == []
     assert ret == 1
@@ -95,112 +90,75 @@ def test_export_empty_serverdir(tmpdir, capfd, monkeypatch, use_option):
     assert ("The path '%s' contains no devpi-server data" % empty) in err
 
 
-@pytest.mark.parametrize("use_option", (False, True))
-def test_export_import(tmpdir, capfd, monkeypatch, use_option):
+def test_export_import(tmpdir, capfd, monkeypatch):
+    from devpi_server.importexport import export
     from devpi_server.importexport import import_
     from devpi_server.init import init
-    from devpi_server.main import main
     monkeypatch.setattr("devpi_server.main.configure_logging", lambda a: None)
     clean = tmpdir.join("clean").ensure(dir=True)
-    if use_option:
-        ret = main(argv=[
-            "devpi-server",
-            "--init",
-            "--serverdir", clean.strpath])
-    else:
-        ret = init(argv=[
-            "devpi-init",
-            "--serverdir", clean.strpath])
+    ret = init(argv=[
+        "devpi-init",
+        "--serverdir", clean.strpath])
     assert ret == 0
-    export = tmpdir.join("export")
-    ret = main([
-        "devpi-server",
+    export_dir = tmpdir.join("export")
+    ret = export(argv=[
+        "devpi-export",
         "--serverdir", clean.strpath,
-        "--export", export.strpath])
+        export_dir.strpath])
     assert ret == 0
     import_dir = tmpdir.join("import")
-    if use_option:
-        ret = main(argv=[
-            "devpi-server",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            "--import", export.strpath])
-    else:
-        ret = import_(argv=[
-            "devpi-import",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            export.strpath])
+    ret = import_(argv=[
+        "devpi-import",
+        "--serverdir", import_dir.strpath,
+        "--no-events",
+        export_dir.strpath])
     assert ret == 0
     out, err = capfd.readouterr()
-    assert os.listdir(clean.strpath) == os.listdir(import_dir.strpath)
+    assert sorted(os.listdir(clean.strpath)) == sorted(os.listdir(import_dir.strpath))
     assert 'import_all: importing finished' in out
     assert err == ''
 
 
-@pytest.mark.parametrize("use_option", (False, True))
-def test_export_import_no_root_pypi(tmpdir, capfd, monkeypatch, use_option):
+def test_export_import_no_root_pypi(tmpdir, capfd, monkeypatch):
+    from devpi_server.importexport import export
     from devpi_server.importexport import import_
     from devpi_server.init import init
-    from devpi_server.main import main
     monkeypatch.setattr("devpi_server.main.configure_logging", lambda a: None)
     clean = tmpdir.join("clean").ensure(dir=True)
-    if use_option:
-        ret = main(argv=[
-            "devpi-server",
-            "--init",
-            "--serverdir", clean.strpath,
-            "--no-root-pypi"])
-    else:
-        ret = init(argv=[
-            "devpi-init",
-            "--serverdir", clean.strpath,
-            "--no-root-pypi"])
+    ret = init(argv=[
+        "devpi-init",
+        "--serverdir", clean.strpath,
+        "--no-root-pypi"])
     assert ret == 0
-    export = tmpdir.join("export")
-    ret = main([
+    export_dir = tmpdir.join("export")
+    ret = export(argv=[
         "devpi-server",
         "--serverdir", clean.strpath,
-        "--export", export.strpath])
+        export_dir.strpath])
     assert ret == 0
     # first we test regular import
     import_dir = tmpdir.join("import")
-    if use_option:
-        ret = main(argv=[
-            "devpi-server",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            "--import", export.strpath])
-    else:
-        ret = import_(argv=[
-            "devpi-import",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            export.strpath])
+    ret = import_(argv=[
+        "devpi-import",
+        "--serverdir", import_dir.strpath,
+        "--no-events",
+        export_dir.strpath])
     assert ret == 0
     out, err = capfd.readouterr()
-    assert os.listdir(clean.strpath) == os.listdir(import_dir.strpath)
+    assert sorted(os.listdir(clean.strpath)) == sorted(os.listdir(import_dir.strpath))
     assert 'import_all: importing finished' in out
     assert err == ''
     # now we add --no-root-pypi
     import_dir.remove()
-    if use_option:
-        ret = main(argv=[
-            "devpi-server",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            "--no-root-pypi",
-            "--import", export.strpath])
-    else:
-        ret = import_(argv=[
-            "devpi-import",
-            "--serverdir", import_dir.strpath,
-            "--no-events",
-            "--no-root-pypi",
-            export.strpath])
+    ret = import_(argv=[
+        "devpi-import",
+        "--serverdir", import_dir.strpath,
+        "--no-events",
+        "--no-root-pypi",
+        export_dir.strpath])
     assert ret == 0
     out, err = capfd.readouterr()
-    assert os.listdir(clean.strpath) == os.listdir(import_dir.strpath)
+    assert sorted(os.listdir(clean.strpath)) == sorted(os.listdir(import_dir.strpath))
     assert 'import_all: importing finished' in out
     assert err == ''
 
@@ -235,84 +193,61 @@ class TestIndexTree:
 
 
 class TestImportExport:
-    @pytest.fixture(params=[False, True])  # switch between cmds and options
+    @pytest.fixture()
     def makeimpexp(self, request, makemapp, gentmp, storage_info):
         class ImpExp:
             def __init__(self, options=()):
+                from devpi_server.main import set_state_version
                 self.exportdir = gentmp()
                 self.testdatadir = py.path.local(
                     pkg_resources.resource_filename(
                         'test_devpi_server', 'importexportdata'))
-                if request.param:
-                    self.mapp1 = makemapp(
-                        options=("--export", self.exportdir) + options)
-                else:
-                    from devpi_server.main import set_state_version
-                    self.mapp1 = makemapp()
-                    set_state_version(self.mapp1.xom.config)
-                    self.options = options
+                self.mapp1 = makemapp()
+                set_state_version(self.mapp1.xom.config)
+                self.options = options
 
             def export(self, initnodeinfo=True):
-                if request.param:
-                    if initnodeinfo:
-                        self.mapp1.xom.config.init_nodeinfo()
-                    assert self.mapp1.xom.main() == 0
-                else:
-                    from devpi_server.importexport import export
-                    argv = [
-                        "devpi-export",
-                        "--serverdir", self.mapp1.xom.config.serverdir]
-                    argv.extend(self.options)
-                    argv.extend(["--storage", storage_info["name"]])
-                    argv.append(self.exportdir)
-                    assert export(argv=argv) == 0
+                from devpi_server.importexport import export
+                argv = [
+                    "devpi-export",
+                    "--serverdir", self.mapp1.xom.config.serverdir]
+                argv.extend(self.options)
+                argv.extend(["--storage", storage_info["name"]])
+                argv.append(self.exportdir)
+                assert export(argv=argv) == 0
 
             def import_testdata(self, name, options=()):
+                from devpi_server.importexport import import_
                 path = self.testdatadir.join(name).strpath
-                if request.param:
-                    mapp = makemapp(
-                        options=("--import", path) + options)
-                    mapp.xom.config.init_nodeinfo()
-                    assert mapp.xom.main() == 0
-                else:
-                    from devpi_server.importexport import import_
-                    serverdir = gentmp()
-                    argv = [
-                        "devpi-import",
-                        "--no-events",
-                        "--serverdir", serverdir]
-                    argv.extend(options)
-                    argv.extend(["--storage", storage_info["name"]])
-                    argv.append(path)
-                    assert import_(argv=argv) == 0
-                    mapp = makemapp(options=["--serverdir", serverdir])
+                serverdir = gentmp()
+                argv = [
+                    "devpi-import",
+                    "--no-events",
+                    "--serverdir", serverdir]
+                argv.extend(options)
+                argv.extend(["--storage", storage_info["name"]])
+                argv.append(path)
+                assert import_(argv=argv) == 0
+                mapp = makemapp(options=["--serverdir", serverdir])
                 return mapp
 
             def new_import(self, options=(), plugin=None):
-                if request.param:
-                    mapp2 = makemapp(
-                        options=("--import", str(self.exportdir)) + options)
-                    if plugin is not None:
-                        mapp2.xom.config.pluginmanager.register(plugin)
-                    mapp2.xom.config.init_nodeinfo()
-                    assert mapp2.xom.main() == 0
-                else:
-                    from devpi_server.config import get_pluginmanager
-                    from devpi_server.importexport import import_
-                    serverdir = gentmp()
-                    argv = [
-                        "devpi-import",
-                        "--no-events",
-                        "--serverdir", serverdir]
-                    argv.extend(options)
-                    argv.extend(["--storage", storage_info["name"]])
-                    argv.append(self.exportdir)
-                    pm = get_pluginmanager()
-                    pm.register(plugin)
-                    assert import_(pluginmanager=pm, argv=argv) == 0
-                    mapp2 = makemapp(options=["--serverdir", serverdir])
-                    if plugin is not None:
-                        mapp2.xom.config.pluginmanager.register(plugin)
+                from devpi_server.config import get_pluginmanager
+                from devpi_server.importexport import import_
+                serverdir = gentmp()
+                argv = [
+                    "devpi-import",
+                    "--no-events",
+                    "--serverdir", serverdir]
+                argv.extend(options)
+                argv.extend(["--storage", storage_info["name"]])
+                argv.append(self.exportdir)
+                pm = get_pluginmanager()
+                pm.register(plugin)
+                assert import_(pluginmanager=pm, argv=argv) == 0
+                mapp2 = makemapp(options=["--serverdir", serverdir])
+                if plugin is not None:
+                    mapp2.xom.config.pluginmanager.register(plugin)
                 return mapp2
         return ImpExp
 
@@ -320,6 +255,49 @@ class TestImportExport:
     def impexp(self, makeimpexp):
         return makeimpexp()
 
+    def test_created_and_modified_old_data(self, impexp, mock, monkeypatch):
+        from time import strftime
+        import datetime
+        gmtime_mock = mock.Mock()
+        monkeypatch.setattr("devpi_server.model.gmtime", gmtime_mock)
+        import_time = datetime.datetime(2021, 2, 22, 10, 51, 51).timetuple()
+        gmtime_mock.return_value = import_time
+        mapp = impexp.import_testdata('nocreatedmodified')
+        users = mapp.getuserlist()
+        # the import time is used for root
+        assert users["root"]["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", import_time)
+        assert "modified" not in users["root"]
+        # the epoch is used for users
+        assert users["user1"]["created"] == "1970-01-01T00:00:00Z"
+        assert "modified" not in users["user1"]
+
+    def test_created_and_modified_existing_data(self, impexp):
+        mapp = impexp.import_testdata('createdmodified')
+        users = mapp.getuserlist()
+        # the time from the import data is used
+        assert users["root"]["created"] == "2020-01-01T11:11:00Z"
+        assert users["root"]["modified"] == "2020-01-02T12:12:00Z"
+
+    def test_created_and_modified_roundtrip(self, impexp):
+        mapp1 = impexp.mapp1
+        mapp1.create_and_use()
+        first_user = mapp1.api.user
+        mapp1.modify_user(first_user, email='foo@example.com')
+        mapp1.create_and_use()
+        second_user = mapp1.api.user
+        users1 = mapp1.getuserlist()
+        impexp.export()
+        mapp2 = impexp.new_import()
+        users2 = mapp2.getuserlist()
+        assert users1["root"]["created"] == users2["root"]["created"]
+        assert "modified" not in users1["root"]
+        assert "modified" not in users2["root"]
+        assert users1[first_user]["created"] == users2[first_user]["created"]
+        assert users1[first_user]["modified"] == users2[first_user]["modified"]
+        assert users1[second_user]["created"] == users2[second_user]["created"]
+        assert "modified" not in users1[second_user]
+        assert "modified" not in users2[second_user]
+
     def test_importing_multiple_indexes_with_releases(self, impexp):
         mapp1 = impexp.mapp1
         api1 = mapp1.create_and_use()
@@ -473,18 +451,24 @@ class TestImportExport:
             "package",
             '<a href="/package-1.0.zip" />\n'
             '<a href="/package-1.1.zip#sha256=a665a45920422f9d417e4867efdc4fb8a04a1f3fff1fa07e998e86f7f7a27ae3" />\n'
-            + '<a href="/package-2.0.zip#sha256=b3a8e0e1f9ab1bfe3a36f231f676f78bb30a519d2b21e6c530c0eee8ebb4a5d0" data-requires-python="&gt;=3.5" />')
+            '<a href="/package-1.2.zip#sha256=b3a8e0e1f9ab1bfe3a36f231f676f78bb30a519d2b21e6c530c0eee8ebb4a5d0" data-yanked="" />\n'
+            '<a href="/package-2.0.zip#sha256=35a9e381b1a27567549b5f8a6f783c167ebf809f1c4d6a9e367240484d8ce281" data-requires-python="&gt;=3.5" />')
         pypistage.mock_extfile("/package-1.1.zip", b"123")
-        pypistage.mock_extfile("/package-2.0.zip", b"456")
+        pypistage.mock_extfile("/package-1.2.zip", b"456")
+        pypistage.mock_extfile("/package-2.0.zip", b"789")
         r = testapp.get(api.index + "/+simple/package/")
         assert r.status_code == 200
         # fetch some files, so they are included in the dump
-        (_, link1, link2) = sorted(x.attrs['href'] for x in r.html.select('a'))
+        (_, link1, link2, link3) = sorted(
+            (x.attrs['href'] for x in r.html.select('a')),
+            key=lambda x: x.split('/')[-1])
         baseurl = URL(r.request.url)
         r = testapp.get(baseurl.joinpath(link1).url)
         assert r.body == b"123"
         r = testapp.get(baseurl.joinpath(link2).url)
         assert r.body == b"456"
+        r = testapp.get(baseurl.joinpath(link3).url)
+        assert r.body == b"789"
         impexp.export()
         mapp2 = impexp.new_import()
         with mapp2.xom.keyfs.transaction(write=False):
@@ -495,8 +479,9 @@ class TestImportExport:
             links = sorted(get_mutable_deepcopy(
                 stage.get_simplelinks_perstage("package")))
             assert links == [
-                ('package-1.1.zip', 'root/pypi/+f/a66/5a45920422f9d/package-1.1.zip', None),
-                ('package-2.0.zip', 'root/pypi/+f/b3a/8e0e1f9ab1bfe/package-2.0.zip', '>=3.5')]
+                ('package-1.1.zip', 'root/pypi/+f/a66/5a45920422f9d/package-1.1.zip', None, None),
+                ('package-1.2.zip', 'root/pypi/+f/b3a/8e0e1f9ab1bfe/package-1.2.zip', None, True),
+                ('package-2.0.zip', 'root/pypi/+f/35a/9e381b1a27567/package-2.0.zip', '>=3.5', None)]
 
     def test_mirrordata(self, caplog, impexp):
         mapp = impexp.import_testdata('mirrordata')
@@ -746,7 +731,7 @@ class TestImportExport:
         with mapp1.xom.keyfs.transaction(write=True):
             stage = mapp1.xom.model.getstage(api.stagename)
             doccontent = zip_dict({"index.html": "<html><body>Hello"})
-            link1 = stage.store_doczip( name, "1.0", content=doccontent)
+            link1 = stage.store_doczip(name, "1.0", content=doccontent)
 
         impexp.export()
 
@@ -1044,9 +1029,8 @@ class TestImportExport:
 
         impexp.export()
 
-        with pytest.raises(InvalidIndexconfig) as excinfo:
+        with pytest.raises(InvalidIndexconfig, match="uploadtrigger_jenkins"):
             impexp.new_import()
-        assert "uploadtrigger_jenkins" in excinfo.value.args[0][0]
 
     def test_plugin_index_config(self, impexp):
         class Plugin:
