@@ -9,18 +9,25 @@ from devpi_server.readonly import is_deeply_readonly
 notransaction = pytest.mark.notransaction
 
 
-
-@pytest.yield_fixture
+@pytest.fixture
 def keyfs(gentmp, pool, storage):
     keyfs = KeyFS(gentmp(), storage)
     pool.register(keyfs.notifier)
     yield keyfs
 
+
 @pytest.fixture(params=["direct", "a/b/c", ])
 def key(request):
     return request.param
 
 
+@pytest.fixture
+def pool():
+    pool = ThreadPool()
+    yield pool
+    pool.shutdown()
+
+
 class TestKeyFS:
     def test_get_non_existent(self, keyfs):
         key = keyfs.add_key("NAME", "somekey", dict)
@@ -187,6 +194,7 @@ class Test_addkey_combinations:
         assert attr.exists()
         assert attr.get() == val
 
+
 class TestKey:
     def test_addkey_type_mismatch(self, keyfs):
         dictkey = keyfs.add_key("NAME1", "some", dict)
@@ -273,14 +281,17 @@ def test_trans_get_not_modify(keyfs, type, val, monkeypatch):
         assert attr.get() == val
     # make sure keyfs doesn't write during the transaction and its commit
     orig_write = py.path.local.write
+
     def write_checker(path, content):
         assert not path.endswith(attr.relpath)
         orig_write(path, content)
+
     monkeypatch.setattr(py.path.local, "write", write_checker)
     with keyfs.transaction():
         x = attr.get()
     assert x == val
 
+
 @notransaction
 class TestTransactionIsolation:
     def test_cannot_write_on_read_trans(self, keyfs):
@@ -291,11 +302,11 @@ class TestTransactionIsolation:
         with pytest.raises(keyfs.ReadOnly):
             tx_1.delete(key)
 
-    def test_serialized_writing(self, queue, keyfs):
+    def test_serialized_writing(self, TimeoutQueue, keyfs):
         import threading
-        from test_devpi_server.conftest import TimeoutQueue
         q1 = TimeoutQueue()
         q2 = TimeoutQueue()
+
         def trans1():
             with keyfs.transaction(write=True):
                 q1.put("write1")
@@ -318,12 +329,12 @@ class TestTransactionIsolation:
         for i in range(NUM):
             assert q1.get() == "write2"
 
-    def test_reading_while_writing(self, queue, keyfs):
+    def test_reading_while_writing(self, TimeoutQueue, keyfs):
         import threading
-        from test_devpi_server.conftest import TimeoutQueue
         q1 = TimeoutQueue()
         q2 = TimeoutQueue()
         q3 = TimeoutQueue()
+
         def trans1():
             with keyfs.transaction(write=True):
                 q1.put("write1")
@@ -347,7 +358,6 @@ class TestTransactionIsolation:
         q2.put("1")
         assert q1.get() == "write1b"
 
-
     def test_concurrent_tx_sees_original_value_on_write(self, keyfs):
         D = keyfs.add_key("NAME", "hello", dict)
         tx_1 = Transaction(keyfs, write=True)
@@ -475,7 +485,7 @@ class TestTransactionIsolation:
         with keyfs.transaction() as tx:
             changes = tx.conn.get_changes(0)
         new_keyfs.import_changes(0, changes)
-        assert l[0][1:] == (new_keyfs.NAME(name="world"), {1:1}, -1)
+        assert l[0][2:] == (new_keyfs.NAME(name="world"), {1:1}, -1)
 
     def test_import_changes_subscriber_error(self, keyfs, storage, tmpdir):
         pkey = keyfs.add_key("NAME", "hello/{name}", dict)
@@ -496,6 +506,7 @@ class TestTransactionIsolation:
         with keyfs.transaction() as tx:
             assert tx.conn.get_raw_changelog_entry(10000) is None
 
+
 @notransaction
 class TestDeriveKey:
     def test_direct_from_file(self, keyfs):
@@ -538,6 +549,10 @@ class TestDeriveKey:
             assert key.params == params
 
 
+@pytest.fixture
+def queue(TimeoutQueue):
+    return TimeoutQueue()
+
 
 @notransaction
 class TestSubscriber:
@@ -556,9 +571,11 @@ class TestSubscriber:
 
     def test_change_subscription_fails(self, keyfs, queue, pool):
         key1 = keyfs.add_key("NAME1", "hello", int)
+
         def failing(event):
             queue.put("willfail")
             0/0
+
         keyfs.notifier.on_key_change(key1, failing)
         keyfs.notifier.on_key_change(key1, queue.put)
         pool.start()
@@ -620,9 +637,11 @@ class TestSubscriber:
     def test_wait_event_serial(self, keyfs, pool, queue, meth):
         pkey = keyfs.add_key("NAME1", "{name}", int)
         key = pkey(name="hello")
+
         class T:
             def __init__(self, num):
                 self.num = num
+
             def thread_run(self):
                 if meth == "wait_event_serial":
                     keyfs.notifier.wait_event_serial(self.num)
@@ -647,10 +666,12 @@ class TestSubscriber:
     def test_wait_tx_async(self, keyfs, pool, queue):
         # start a thread which waits for the next serial
         wait_serial = keyfs.get_next_serial()
+
         class T:
             def thread_run(self):
                 ret = keyfs.wait_tx_serial(wait_serial, recheck=0.01)
                 queue.put(ret)
+
         pool.register(T())
         pool.start()
 
