@@ -5,14 +5,12 @@ from .functional import TestMirrorIndexThings as BaseTestMirrorIndexThings
 import pytest
 
 
-@pytest.yield_fixture
-def mapp(makemapp, nginx_host_port):
-    from devpi_server.replica import ReplicaThread
-    app = makemapp(options=['--master', 'http://%s:%s' % nginx_host_port])
-    rt = ReplicaThread(app.xom)
-    app.xom.replica_thread = rt
-    app.xom.thread_pool.register(rt)
-    app.xom.thread_pool.start_one(rt)
+@pytest.fixture
+def mapp(makemapp, nginx_host_port, secretfile):
+    app = makemapp(options=[
+        '--master', 'http://%s:%s' % nginx_host_port,
+        '--secretfile', secretfile.strpath])
+    app.xom.thread_pool.start_one(app.xom.replica_thread)
     try:
         yield app
     finally:
