@@ -1,17 +1,14 @@
 # -*- coding: utf-8 -*-
 import hashlib
-import json
 import os
 import pytest
 from devpi_server.log import thread_pop_log
 from devpi_server.fileutil import loads
+from devpi_server.keyfs import MissingFileException
 from devpi_server.log import threadlog, thread_push_log
 from devpi_server.replica import H_EXPECTED_MASTER_ID, H_MASTER_UUID
 from devpi_server.replica import H_REPLICA_UUID, H_REPLICA_OUTSIDE_URL
-from devpi_server.replica import FileReplicationError
-from devpi_server.replica import ImportFileReplica
 from devpi_server.replica import MasterChangelogRequest
-from devpi_server.replica import ReplicaThread
 from devpi_server.replica import proxy_view_to_master
 from devpi_server.views import iter_remote_file_replica
 from pyramid.httpexceptions import HTTPNotFound
@@ -20,6 +17,14 @@ from pyramid.httpexceptions import HTTPNotFound
 pytestmark = [pytest.mark.notransaction]
 
 
+@pytest.fixture
+def auth_serializer(xom):
+    import itsdangerous
+
+    return itsdangerous.TimedSerializer(
+        xom.config.get_replica_secret())
+
+
 @pytest.fixture
 def replica_pypistage(devpiserver_makepypistage, replica_xom):
     return devpiserver_makepypistage(replica_xom)
@@ -40,10 +45,12 @@ class TestChangelog:
     replica_url = "http://qwe"
 
     @pytest.fixture(params=[False, True])
-    def reqchangelog(self, request, testapp):
+    def reqchangelog(self, request, auth_serializer, testapp, xom):
         def reqchangelog(serial):
+            token = auth_serializer.dumps(self.replica_uuid)
             req_headers = {H_REPLICA_UUID: self.replica_uuid,
-                           H_REPLICA_OUTSIDE_URL: self.replica_url}
+                           H_REPLICA_OUTSIDE_URL: self.replica_url,
+                           str('Authorization'): 'Bearer %s' % token}
             url = "/+changelog/%s" % serial
             use_multi_endpoint = request.param
             if use_multi_endpoint:
@@ -89,9 +96,16 @@ class TestChangelog:
             serial = mcr._wait_for_serial(xom.keyfs.get_current_serial())
         assert serial == 1
 
-    def test_master_id_mismatch(self, testapp):
-        testapp.xget(400, "/+changelog/0", headers={H_EXPECTED_MASTER_ID:str("123")})
-        r = testapp.xget(200, "/+changelog/0", headers={H_EXPECTED_MASTER_ID: ''})
+    def test_master_id_mismatch(self, auth_serializer, testapp):
+        token = auth_serializer.dumps(self.replica_uuid)
+        testapp.xget(400, "/+changelog/0", headers={
+            H_REPLICA_UUID: self.replica_uuid,
+            H_EXPECTED_MASTER_ID:str("123"),
+            str('Authorization'): 'Bearer %s' % token})
+        r = testapp.xget(200, "/+changelog/0", headers={
+            H_REPLICA_UUID: self.replica_uuid,
+            H_EXPECTED_MASTER_ID: '',
+            str('Authorization'): 'Bearer %s' % token})
         assert r.headers[H_MASTER_UUID]
         del testapp.headers[H_EXPECTED_MASTER_ID]
         testapp.xget(400, "/+changelog/0")
@@ -102,10 +116,12 @@ class TestMultiChangelog:
     replica_url = "http://qwe"
 
     @pytest.fixture
-    def reqchangelogs(self, request, testapp):
+    def reqchangelogs(self, request, auth_serializer, testapp):
         def reqchangelogs(serial):
+            token = auth_serializer.dumps(self.replica_uuid)
             req_headers = {H_REPLICA_UUID: self.replica_uuid,
-                           H_REPLICA_OUTSIDE_URL: self.replica_url}
+                           H_REPLICA_OUTSIDE_URL: self.replica_url,
+                           str('Authorization'): 'Bearer %s' % token}
             url = "/+changelog/%s-" % serial
             return testapp.get(url, expect_errors=False, headers=req_headers)
         return reqchangelogs
@@ -145,13 +161,12 @@ def get_raw_changelog_entry(xom, serial):
     with xom.keyfs._storage.get_connection() as conn:
         return conn.get_raw_changelog_entry(serial)
 
+
 class TestReplicaThread:
     @pytest.fixture
     def rt(self, makexom):
         xom = makexom(["--master=http://localhost"])
-        rt = ReplicaThread(xom)
-        xom.thread_pool.register(rt)
-        return rt
+        return xom.replica_thread
 
     @pytest.fixture
     def mockchangelog(self, reqmock):
@@ -165,8 +180,10 @@ class TestReplicaThread:
             headers.setdefault("x-devpi-serial", str(2))
             if headers["x-devpi-serial"] is None:
                 del headers["x-devpi-serial"]
-            reqmock.mockresponse("http://localhost/+changelog/%s" % num,
-                                 code=code, data=data, headers=headers)
+            url = "http://localhost/+changelog/%s" % num
+            if num == 0:
+                url = url + '?initial_fetch'
+            reqmock.mockresponse(url, code=code, data=data, headers=headers)
         return mockchangelog
 
     def test_thread_run_fail(self, rt, mockchangelog, caplog):
@@ -189,7 +206,7 @@ class TestReplicaThread:
         data = get_raw_changelog_entry(xom, 0)
         mockchangelog(0, code=200, data=data)
         # get the result
-        orig_req = reqmock.url2reply[("http://localhost/+changelog/0", None)]
+        orig_req = reqmock.url2reply[("http://localhost/+changelog/0?initial_fetch", None)]
         # setup so the first attempt fails, then the second succeeds
         reqmock.url2reply = mock.Mock()
         reqmock.url2reply.get.side_effect = [socket.error(), orig_req]
@@ -282,8 +299,10 @@ class TestReplicaThread:
 
     def test_thread_run_try_again(self, rt, mockchangelog, caplog):
         l = [1]
+
         def exit_if_shutdown():
             l.pop()
+
         rt.thread.exit_if_shutdown = exit_if_shutdown
         mockchangelog(0, code=202)
         with pytest.raises(IndexError):
@@ -378,7 +397,13 @@ class TestProxyViewToMaster:
         assert 'foo' not in response.headers
         assert 'keep-alive' not in response.headers
 
+
 def replay(xom, replica_xom, events=True):
+    if replica_xom.replica_thread.replica_in_sync_at is None:
+        # allow on_import to run right away, so we don't need to rely
+        # on the initial import thread for tests
+        replica_xom.replica_thread.replica_in_sync_at = 0
+
     threadlog.info("test: replaying replica")
     for serial in range(replica_xom.keyfs.get_next_serial(),
                         xom.keyfs.get_next_serial()):
@@ -391,6 +416,7 @@ def replay(xom, replica_xom, events=True):
 
     # replay notifications
     if events:
+        replica_xom.replica_thread.wait()
         noti_thread = replica_xom.keyfs.notifier
         event_serial = noti_thread.read_event_serial()
         thread_push_log("NOTI")
@@ -402,15 +428,17 @@ def replay(xom, replica_xom, events=True):
 
 
 @pytest.fixture
-def make_replica_xom(makexom):
+def make_replica_xom(makexom, secretfile):
     def make_replica_xom(options=()):
-        from devpi_server.replica import ReplicationErrors
-        replica_xom = makexom(["--master", "http://localhost"] + list(options))
-        keyfs = replica_xom.keyfs
-        replica_xom.errors = ReplicationErrors(replica_xom.config.serverdir)
-        for key in (keyfs.STAGEFILE, keyfs.PYPIFILE_NOMD5):
-            keyfs.subscribe_on_import(
-                key, ImportFileReplica(replica_xom, replica_xom.errors))
+        replica_xom = makexom([
+            "--master", "http://localhost",
+            "--file-replication-threads", "1",
+            "--secretfile", secretfile.strpath] + list(options))
+        # shorten error delay for tests
+        replica_xom.replica_thread.shared_data.ERROR_QUEUE_MAX_DELAY = 0.1
+        replica_xom.thread_pool.start_one(replica_xom.replica_thread)
+        replica_xom.thread_pool.start_one(
+            replica_xom.replica_thread.file_replication_threads[0])
         return replica_xom
     return make_replica_xom
 
@@ -434,7 +462,8 @@ class TestUseExistingFiles:
             '--replica-file-search-path', existing_base.strpath])
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
-        replay(xom, replica_xom)
+        replay(xom, replica_xom, events=False)
+        replica_xom.replica_thread.wait()
         assert len(caplog.getrecords('checking existing file')) == 1
 
     @pytest.mark.storage_with_filesystem
@@ -465,7 +494,7 @@ class TestUseExistingFiles:
         # check the number of links of the file
         assert existing_path.stat().nlink == 2
 
-    def test_use_existing_files_bad_data(self, caplog, httpget, make_replica_xom, mapp, monkeypatch, tmpdir, xom):
+    def test_use_existing_files_bad_data(self, caplog, make_replica_xom, mapp, monkeypatch, patch_reqsessionmock, tmpdir, xom):
         # this will be the folder to find existing files in the replica
         existing_base = tmpdir.join('existing').ensure_dir()
         # prepare data on master
@@ -474,7 +503,6 @@ class TestUseExistingFiles:
         mapp.upload_file_pypi("hello-1.0.zip", content1, "hello", "1.0")
         # get the path of the release
         (path,) = mapp.get_release_paths('hello')
-        httpget.mockresponse("http://localhost" + path, content=content1)
         # create the file
         existing_path = existing_base.join(path)
         existing_path.dirpath().ensure_dir()
@@ -482,9 +510,13 @@ class TestUseExistingFiles:
         # create the replica with the path to existing files
         replica_xom = make_replica_xom(options=[
             '--replica-file-search-path', existing_base.strpath])
+        (frthread,) = replica_xom.replica_thread.file_replication_threads
+        frt_reqmock = patch_reqsessionmock(frthread.session)
+        frt_reqmock.mockresponse("http://localhost" + path, 200, data=content1)
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
-        replay(xom, replica_xom)
+        replay(xom, replica_xom, events=False)
+        replica_xom.replica_thread.wait()
         assert len(caplog.getrecords('checking existing file')) == 1
         assert len(caplog.getrecords('sha256 mismatch, got ec7d057f450dc963f15978af98b9cdda64aca6751c677e45d4a358fe103dc05b')) == 1
         with replica_xom.keyfs.transaction(write=False):
@@ -494,7 +526,7 @@ class TestUseExistingFiles:
     @pytest.mark.storage_with_filesystem
     @pytest.mark.skipif(not hasattr(os, 'link'),
                         reason="OS doesn't support hard links")
-    def test_hardlink_bad_data(self, caplog, httpget, make_replica_xom, mapp, monkeypatch, tmpdir, xom):
+    def test_hardlink_bad_data(self, caplog, make_replica_xom, mapp, monkeypatch, patch_reqsessionmock, tmpdir, xom):
         # this will be the folder to find existing files in the replica
         existing_base = tmpdir.join('existing').ensure_dir()
         # prepare data on master
@@ -503,7 +535,6 @@ class TestUseExistingFiles:
         mapp.upload_file_pypi("hello-1.0.zip", content1, "hello", "1.0")
         # get the path of the release
         (path,) = mapp.get_release_paths('hello')
-        httpget.mockresponse("http://localhost" + path, content=content1)
         # create the file
         existing_path = existing_base.join(path)
         existing_path.dirpath().ensure_dir()
@@ -513,6 +544,9 @@ class TestUseExistingFiles:
         replica_xom = make_replica_xom(options=[
             '--replica-file-search-path', existing_base.strpath,
             '--hard-links'])
+        (frthread,) = replica_xom.replica_thread.file_replication_threads
+        frt_reqmock = patch_reqsessionmock(frthread.session)
+        frt_reqmock.mockresponse("http://localhost" + path, 200, data=content1)
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
         replay(xom, replica_xom)
@@ -541,7 +575,6 @@ class TestFileReplication:
             with replica_xom.keyfs.transaction():
                 replica_xom.keyfs.restart_as_write_transaction()
 
-
     def test_transaction_api(self, replica_xom, xom):
         with xom.keyfs.transaction(write=True):
             xom.model.create_user("hello", "pass")
@@ -573,17 +606,17 @@ class TestFileReplication:
         with xom.keyfs.transaction(write=True):
             entry.file_set_content(content1)
             assert entry.file_exists()
+            assert entry.last_modified is not None
 
         # first we try to return something wrong
         master_url = replica_xom.config.master_url
         master_file_path = master_url.joinpath(entry.relpath).url
-        xom.httpget.mockresponse(master_file_path, code=200, content=b'13')
-        replay(xom, replica_xom)
-        assert list(replica_xom.errors.errors.keys()) == [
+        reqmock.mockresponse(master_file_path, code=200, data=b'13')
+        replay(xom, replica_xom, events=False)
+        replica_xom.replica_thread.wait(error_queue=True)
+        replication_errors = replica_xom.replica_thread.shared_data.errors
+        assert list(replication_errors.errors.keys()) == [
             'root/pypi/+f/5d4/1402abc4b2a76/pytest-1.8.zip']
-        with replica_xom.errors.errorsfn.open() as f:
-            persisted_errors = json.load(f)
-        assert persisted_errors == replica_xom.errors.errors
         with replica_xom.keyfs.transaction():
             assert not r_entry.file_exists()
             assert not replica_xom.config.serverdir.join(r_entry._storepath).exists()
@@ -592,12 +625,9 @@ class TestFileReplication:
         with xom.keyfs.transaction(write=True):
             # trigger a change
             entry.last_modified = 'Fri, 09 Aug 2019 13:15:02 GMT'
-        xom.httpget.mockresponse(master_file_path, code=200, content=content1)
+        reqmock.mockresponse(master_file_path, code=200, data=content1)
         replay(xom, replica_xom)
-        assert replica_xom.errors.errors == {}
-        with replica_xom.errors.errorsfn.open() as f:
-            persisted_errors = json.load(f)
-        assert persisted_errors == replica_xom.errors.errors
+        assert replication_errors.errors == {}
         with replica_xom.keyfs.transaction():
             assert r_entry.file_exists()
             assert r_entry.file_get_content() == content1
@@ -641,7 +671,9 @@ class TestFileReplication:
             r_entry = replica_xom.filestore.get_file_entry(entry.relpath)
             assert not r_entry.file_exists()
 
-    def test_fetch_pypi_nomd5(self, gen, reqmock, xom, replica_xom):
+    def test_fetch_pypi_nomd5(self, gen, patch_reqsessionmock, reqmock, xom, replica_xom):
+        (frthread,) = replica_xom.replica_thread.file_replication_threads
+        frt_reqmock = patch_reqsessionmock(frthread.session)
         replay(xom, replica_xom)
         content1 = b'hello'
         link = gen.pypi_package_link("some-1.8.zip", md5=False)
@@ -663,19 +695,35 @@ class TestFileReplication:
         master_url = replica_xom.config.master_url
         master_file_path = master_url.joinpath(entry.relpath).url
         # simulate some 500 master server error
-        replica_xom.httpget.mockresponse(master_file_path, status_code=500,
-                                         content=b'')
-        with pytest.raises(FileReplicationError) as e:
+        frt_reqmock.mockresponse(
+            master_file_path, code=500, data=b'')
+        with pytest.raises(MissingFileException) as e:
+            # the event handling will stop with an exception
             replay(xom, replica_xom)
-        assert str(e.value) == 'FileReplicationError with http://localhost/root/pypi/+e/https_pypi.org_package_some/some-1.8.zip, code=500, reason=Internal Server Error, relpath=root/pypi/+e/https_pypi.org_package_some/some-1.8.zip, message=failed'
+        assert str(e.value) == "missing file 'root/pypi/+e/https_pypi.org_package_some/some-1.8.zip' at serial 2"
+        assert replica_xom.keyfs.get_current_serial() == xom.keyfs.get_current_serial()
+        # event handling hasn't progressed
+        assert replica_xom.keyfs.notifier.read_event_serial() == 1
+        # we also got an error entry
+        replication_errors = replica_xom.replica_thread.shared_data.errors
+        assert list(replication_errors.errors.keys()) == [
+            'root/pypi/+e/https_pypi.org_package_some/some-1.8.zip']
 
         # now get the real thing
-        replica_xom.httpget.mockresponse(master_file_path, status_code=200,
-                                         content=content1)
-        replay(xom, replica_xom)
+        frt_reqmock.mockresponse(
+            master_file_path, code=200, data=content1)
+        # wait for the error queue to clear
+        replica_xom.replica_thread.wait(error_queue=True)
+        # there should be no errors anymore
+        assert replication_errors.errors == {}
+        # and the file should exist
         with replica_xom.keyfs.transaction():
             assert r_entry.file_exists()
             assert r_entry.file_get_content() == content1
+        # and the event handling should continue
+        replay(xom, replica_xom)
+        # event handling has continued
+        assert replica_xom.keyfs.notifier.read_event_serial() == replica_xom.keyfs.get_current_serial()
 
     def test_cache_remote_file_fails(self, xom, replica_xom, gen,
                                      pypistage, monkeypatch, reqmock):
@@ -691,7 +739,7 @@ class TestFileReplication:
         with replica_xom.keyfs.transaction():
             entry = replica_xom.filestore.get_file_entry(entry.relpath)
             url = replica_xom.config.master_url.joinpath(entry.relpath).url
-            pypistage.httpget.mockresponse(url, status_code=500)
+            pypistage.xom.httpget.mockresponse(url, status_code=500)
             with pytest.raises(BadGateway) as e:
                 for part in iter_remote_file_replica(replica_xom, entry):
                     pass
@@ -718,20 +766,29 @@ class TestFileReplication:
                 "content-type": ("application/zip", None)}
             entry = replica_xom.filestore.get_file_entry(entry.relpath)
             url = replica_xom.config.master_url.joinpath(entry.relpath).url
-            pypistage.httpget.mockresponse(url, status_code=500)
-            pypistage.httpget.mockresponse(
+            pypistage.xom.httpget.mockresponse(url, status_code=500)
+            pypistage.xom.httpget.mockresponse(
                 entry.url, headers=headers, content=b'123')
             result = iter_remote_file_replica(replica_xom, entry)
             headers = next(result)
+            # there should be one get
+            (call_log_entry,) = [
+                x for x in pypistage.xom.httpget.call_log
+                if x['url'] == url]
+            # and it should have an authorization header
+            assert call_log_entry['extra_headers']['Authorization'].startswith('Bearer')
+            # and UUID header
+            assert call_log_entry['extra_headers'][H_REPLICA_UUID]
             assert headers['content-length'] == '3'
             assert b''.join(result) == b'123'
 
     def test_checksum_mismatch(self, xom, replica_xom, gen, maketestapp,
-                               makemapp, reqmock):
+                               makemapp, patch_reqsessionmock):
         # this test might seem to be doing the same as test_fetch above, but
         # test_fetch creates a new transaction for the same file, which doesn't
         # happen 'in real life'™
-        from devpi_server.replica import ReplicationErrors
+        (frthread,) = replica_xom.replica_thread.file_replication_threads
+        frt_reqmock = patch_reqsessionmock(frthread.session)
         app = maketestapp(xom)
         mapp = makemapp(app)
         api = mapp.create_and_use()
@@ -743,10 +800,11 @@ class TestFileReplication:
         (path,) = mapp.get_release_paths('hello')
         file_relpath = '+files' + path
         master_file_url = master_url.joinpath(path).url
-        replica_xom.httpget.mockresponse(master_file_url, code=200, content=b'13')
-        replay(xom, replica_xom)
+        frt_reqmock.mockresponse(master_file_url, code=200, data=b'13')
+        replay(xom, replica_xom, events=False)
+        replica_xom.replica_thread.wait()
         assert xom.keyfs.get_current_serial() == replica_xom.keyfs.get_current_serial()
-        replication_errors = ReplicationErrors(replica_xom.config.serverdir)
+        replication_errors = replica_xom.replica_thread.shared_data.errors
         assert list(replication_errors.errors.keys()) == [
             '%s/+f/d0b/425e00e15a0d3/hello-1.0.zip' % api.stagename]
         # the master and replica are in sync, so getting the file on the
@@ -755,7 +813,6 @@ class TestFileReplication:
                    "last-modified": "Thu, 25 Nov 2010 20:00:27 GMT",
                    "content-type": "application/zip",
                    "X-DEVPI-SERIAL": str(xom.keyfs.get_current_serial())}
-        reqmock.mockresponse(master_file_url, code=200, headers=headers)
         replica_xom.httpget.mockresponse(master_file_url, code=200, content=content1, headers=headers)
         with replica_xom.keyfs.transaction(write=False) as tx:
             assert not tx.conn.io_file_exists(file_relpath)
@@ -764,20 +821,21 @@ class TestFileReplication:
         assert r.body == content1
         with replica_xom.keyfs.transaction(write=False) as tx:
             assert tx.conn.io_file_exists(file_relpath)
-        replication_errors = ReplicationErrors(replica_xom.config.serverdir)
+        replication_errors = replica_xom.replica_thread.shared_data.errors
         assert list(replication_errors.errors.keys()) == []
 
 
-def test_simplelinks_update_updates_projectname(httpget, monkeypatch,
-        pypistage, replica_pypistage, pypiurls, replica_xom, xom):
-
+def test_simplelinks_update_updates_projectname(pypistage, replica_xom, xom):
+    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
+    replica_xom.thread_pool.start_one(
+        replica_xom.replica_thread.file_replication_threads[0])
     pypistage.mock_simple_projects([])
     pypistage.mock_simple("pytest", pkgver="pytest-1.0.zip")
     with xom.keyfs.transaction():
         assert not pypistage.list_projects_perstage()
 
     with xom.keyfs.transaction():
-        pypistage.get_simplelinks("pytest")
+        assert len(pypistage.get_simplelinks("pytest")) == 1
 
     # replicate including executing events
     replay(xom, replica_xom)
@@ -789,6 +847,10 @@ def test_simplelinks_update_updates_projectname(httpget, monkeypatch,
 
 def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypistage,
                                   pypiurls, replica_xom, xom):
+    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
+    replica_xom.thread_pool.start_one(
+        replica_xom.replica_thread.file_replication_threads[0])
+
     orig_simple = pypiurls.simple
 
     # prepare the data on master
@@ -810,6 +872,14 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
         ret = replica_pypistage.get_releaselinks("pytest")
     assert len(ret) == 1
     assert ret[0].relpath == 'root/pypi/+e/https_pypi.org_pytest/pytest-1.0.zip'
+    # there should be one get
+    (call_log_entry,) = [
+        x for x in pypistage.xom.httpget.call_log
+        if x['url'].startswith(pypiurls.simple)]
+    # and it should have an authorization header
+    assert call_log_entry['extra_headers']['Authorization'].startswith('Bearer')
+    # and UUID header
+    assert call_log_entry['extra_headers'][H_REPLICA_UUID]
 
     # now we change the links and expire the cache
     pypiurls.simple = orig_simple
@@ -819,17 +889,17 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
         pypistage.get_releaselinks("pytest")
     assert xom.keyfs.get_current_serial() > serial
 
-    # we patch wait_tx_serial so we can check and replay
+    # we patch wait_tx_serial so we can check it
+    orig_wait_tx_serial = replica_xom.keyfs.wait_tx_serial
     called = []
+
     def wait_tx_serial(serial, timeout=None):
-        called.append(True)
-        assert xom.keyfs.get_current_serial() == serial
-        assert replica_xom.keyfs.get_current_serial() < serial
-        replay(xom, replica_xom, events=False)
-        assert replica_xom.keyfs.get_current_serial() == serial
-        return True
+        result = orig_wait_tx_serial(serial, timeout=timeout)
+        called.append((serial, result))
+        return result
 
     monkeypatch.setattr(replica_xom.keyfs, 'wait_tx_serial', wait_tx_serial)
+    replay(xom, replica_xom, events=False)
     pypiurls.simple = 'http://localhost:3111/root/pypi/+simple/'
     httpget.mock_simple(
         'pytest',
@@ -841,20 +911,26 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
         r_pypistage = replica_xom.model.getstage("root/pypi")
         r_pypistage.cache_retrieve_times.expire("pytest")
         ret = replica_pypistage.get_releaselinks("pytest")
-    assert called == [True]
+    assert called == [(2, True)]
     replay(xom, replica_xom)
     assert len(ret) == 1
     assert ret[0].relpath == 'root/pypi/+e/https_pypi.org_pytest/pytest-1.1.zip'
 
 
 def test_replicate_deleted_user(mapp, replica_xom):
+    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
+    replica_xom.thread_pool.start_one(
+        replica_xom.replica_thread.file_replication_threads[0])
     mapp.create_and_use("hello/dev")
     content = mapp.makepkg("hello-1.0.tar.gz", b"content", "hello", "1.0")
     mapp.upload_file_pypi("hello-1.0.tar.gz", content, "hello", "1.0")
     mapp.delete_user("hello")
-
     # replicate the state
     replay(mapp.xom, replica_xom)
+    # create again
+    mapp.create_and_use("hello/dev")
+    # replicate the state again
+    replay(mapp.xom, replica_xom)
 
 
 def test_auth_status_master_down(maketestapp, replica_xom, mock):
@@ -867,9 +943,8 @@ def test_auth_status_master_down(maketestapp, replica_xom, mock):
             raise UpstreamError("foo")
         prtm.side_effect = proxy_request_to_master
         r = testapp.get('/+api')
-    assert len(calls) == 1
-    assert calls[0][0][1].url == 'http://localhost/+api'
-    assert r.json['result']['authstatus'] == ['fail', '', []]
+    assert len(calls) == 0
+    assert r.json['result']['authstatus'] == ['noauth', '', []]
 
 
 def test_master_url_auth(makexom, monkeypatch):
@@ -895,3 +970,154 @@ def test_master_url_auth_with_port(makexom):
     replica_xom = makexom(opts=["--master=http://foo:pass@localhost:3140"])
     assert replica_xom.config.master_auth == ("foo", "pass")
     assert replica_xom.config.master_url.url == "http://localhost:3140"
+
+
+def test_replica_user_auth_before_other_plugins(makexom):
+    from devpi_server import replica
+    from devpi_server.config import hookimpl
+    from devpi_server.auth import Auth
+    from pyramid.httpexceptions import HTTPForbidden
+
+    class Plugin:
+        @hookimpl
+        def devpiserver_auth_request(self, request, userdict, username, password):
+            raise RuntimeError("Shouldn't be called")
+
+        @hookimpl
+        def devpiserver_auth_user(self, userdict, username, password):
+            raise RuntimeError("Shouldn't be called")
+
+    plugin = Plugin()
+    # register both plugins, so the above plugin would normally be called first
+    xom = makexom(plugins=[replica, plugin])
+    auth = Auth(xom.model, "qweqwe")
+    with xom.keyfs.transaction(write=False):
+        # because the replica auth has tryfirst our plugin shouldn't be called
+        with pytest.raises(HTTPForbidden):
+            auth._get_auth_status(replica.REPLICA_USER_NAME, '')
+
+
+class TestFileReplicationSharedData:
+    @pytest.fixture
+    def shared_data(self, replica_xom):
+        from devpi_server.replica import FileReplicationSharedData
+        # allow on_import to run right away, so we don't need to rely
+        # on the initial import thread for tests
+        replica_xom.replica_thread.replica_in_sync_at = 0
+        return FileReplicationSharedData(replica_xom)
+
+    def test_mirror_priority(self, shared_data):
+        result = []
+
+        mirror_file = 'root/pypi/+f/3f8/3058ac9076112/pytest-2.0.0.zip'
+        stage_file = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
+        # set the index_types cache to prevent db access
+        shared_data.index_types.put('root/pypi', 'mirror')
+        shared_data.index_types.put('root/dev', 'stage')
+
+        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+            result.append(key)
+        # Regardless of the serial or add order, the stage should come first
+        cases = [
+            ((mirror_file, 0), (stage_file, 0)),
+            ((mirror_file, 1), (stage_file, 0)),
+            ((mirror_file, 0), (stage_file, 1)),
+            ((stage_file, 0), (mirror_file, 0)),
+            ((stage_file, 1), (mirror_file, 0)),
+            ((stage_file, 0), (mirror_file, 1))]
+        for (relpath1, serial1), (relpath2, serial2) in cases:
+            key1 = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath1)
+            key2 = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath2)
+            shared_data.on_import(None, serial1, key1, None, -1)
+            shared_data.on_import(None, serial2, key2, None, -1)
+            assert shared_data.queue.qsize() == 2
+            shared_data.process_next(handler)
+            shared_data.process_next(handler)
+            assert shared_data.queue.qsize() == 0
+            assert result == [stage_file, mirror_file]
+            result.clear()
+
+    @pytest.mark.parametrize("index_type", ["mirror", "stage"])
+    def test_serial_priority(self, index_type, shared_data):
+        relpath = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
+        key = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath)
+        # set the index_types cache to prevent db access
+        shared_data.index_types.put('root/dev', 'stage')
+        result = []
+
+        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+            result.append(serial)
+
+        # Later serials come first
+        shared_data.on_import(None, 1, key, None, -1)
+        shared_data.on_import(None, 100, key, None, -1)
+        shared_data.on_import(None, 10, key, None, -1)
+        assert shared_data.queue.qsize() == 3
+        shared_data.process_next(handler)
+        shared_data.process_next(handler)
+        shared_data.process_next(handler)
+        assert shared_data.queue.qsize() == 0
+        assert result == [100, 10, 1]
+
+    def test_error_queued(self, shared_data):
+        relpath = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
+        key = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath)
+        # set the index_types cache to prevent db access
+        shared_data.index_types.put('root/dev', 'stage')
+
+        next_ts_result = []
+        handler_result = []
+        orig_next_ts = shared_data.next_ts
+
+        def next_ts(delay):
+            next_ts_result.append(delay)
+            return orig_next_ts(delay)
+        shared_data.next_ts = next_ts
+
+        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+            handler_result.append(key)
+            raise ValueError
+
+        # No waiting on empty queues
+        shared_data.QUEUE_TIMEOUT = 0
+        shared_data.on_import(None, 0, key, None, -1)
+        assert shared_data.queue.qsize() == 1
+        assert shared_data.error_queue.qsize() == 0
+        assert next_ts_result == []
+        assert handler_result == []
+        # An exception puts the info into the error queue
+        shared_data.process_next(handler)
+        assert shared_data.queue.qsize() == 0
+        assert shared_data.error_queue.qsize() == 1
+        assert next_ts_result == [11]
+        assert handler_result == [relpath]
+        # Calling again doesn't change anything,
+        # because there is a delay on errors
+        shared_data.process_next(handler)
+        assert shared_data.queue.qsize() == 0
+        assert shared_data.error_queue.qsize() == 1
+        assert next_ts_result == [11]
+        assert handler_result == [relpath]
+        # When removing the delay check, the handler is called again and the
+        # info re-queued with a longer delay
+        shared_data.is_in_future = lambda ts: False
+        shared_data.process_next(handler)
+        assert shared_data.queue.qsize() == 0
+        assert shared_data.error_queue.qsize() == 1
+        assert next_ts_result == [11, 11 * shared_data.ERROR_QUEUE_DELAY_MULTIPLIER]
+        assert handler_result == [relpath, relpath]
+        while 1:
+            # The delay is increased until reaching a maximum
+            shared_data.process_next(handler)
+            delay = next_ts_result[-1]
+            if delay >= shared_data.ERROR_QUEUE_MAX_DELAY:
+                break
+        # then it will stay there
+        shared_data.process_next(handler)
+        delay = next_ts_result[-1]
+        assert delay == shared_data.ERROR_QUEUE_MAX_DELAY
+        # The number of retries should be reasonable.
+        # Needs adjustment in case the ERROR_QUEUE_DELAY_MULTIPLIER
+        # or ERROR_QUEUE_MAX_DELAY is changed
+        assert len(next_ts_result) == 17
+        assert len(handler_result) == 17
